%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Paramètres 
clear;
clf;
N = 100;
K = 2;
NB_KNN = 5;
LINKED = 0;
NB_SUPERVISION = N-1;
NB_CONSTRAINT = 1;
MAX_LOOP = 20;
PROPAGATION2 = false; 
PROPAGATION3 = false; 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialisations
clusters_gt = mod(linspace(1,N,N),K)';
q = zeros(NB_SUPERVISION,5);
qk = zeros(NB_SUPERVISION,5);
qs3 = zeros(NB_SUPERVISION,5);
tps = zeros(NB_SUPERVISION,2);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Loop
for loop=1:MAX_LOOP
    disp('Loop:');
    disp(loop);
    id = 0;
    idk = 0;
    fin = 0;
    ML = sparse(N,N);
    CL = sparse(N,N);
    MLk = sparse(N,N);
    CLk = sparse(N,N);
    MLs3 = sparse(N,N);
    CLs3 = sparse(N,N);
    ql = zeros(NB_SUPERVISION,5);
    qlk = zeros(NB_SUPERVISION,5);
    qls3 = zeros(NB_SUPERVISION,5);
    tpsl = zeros(NB_SUPERVISION,2);
    GetRandomConstraint(0,ML,CL,LINKED,'INIT');
    for i=1:NB_SUPERVISION
        id = id + 1;
        sl = GetRandomConstraint(NB_CONSTRAINT,ML,CL,LINKED);
        if (isempty(sl) && ~fin)
            fin = id-1;
            break;
        end
        [ML,CL,MLadd,CLadd] = Supervise(sl,clusters_gt,ML,CL);
        
        % Appel sans la règle 3 => effet boule de neige (fig 3)
        [MLs3,CLs3,MLadds3,CLadds3] = Supervise(sl,clusters_gt,MLs3,CLs3);
        [MLs3,CLs3,qlids3] = Propagation(MLs3,CLs3,MLadds3,CLadds3,false,false);
        
        % Implémentation matricielle originale => comparaison des temps
        tic;
        Propagation(ML,CL,MLadd,CLadd,K==2,true);
        tps1 = toc;
        
        % Implémentation matric. optimisée => comparaison des temps (fig 1)
        % => détail des règles (fig 2)
        % => effet boule de neige (fig 3)
        tic;
        [ML,CL,qlid] = Propagation(ML,CL,MLadd,CLadd,K==2,false);
        tps2 = toc;
        
        if (id==1)
            ql(id,:) = [qlid,sum(qlid)];
            qls3(id,:) = [qlids3,sum(qlids3)];
            tpsl(id,:) = [tps1,tps2];
        else
            ql(id,:) = ql(id-1,:) + [qlid,sum(qlid)];
            qls3(id,:) = qls3(id-1,:) + [qlids3,sum(qlids3)];
            tpsl(id,:) = tpsl(id-1,:) + [tps1,tps2];
        end
    end
    q = ((loop-1)*q+ql)/loop;
    qs3 = ((loop-1)*qs3+qls3)/loop;
    tps = ((loop-1)*tps+tpsl)/loop;

    GetRandomConstraint(0,MLk,CLk,1,'INIT');
    for i=1:NB_SUPERVISION
        idk = idk + 1;
        slk = GetRandomConstraint(NB_CONSTRAINT,MLk,CLk,1);
        if (isempty(slk) && ~fin)
            fin = idk-1;
            break;
        end
        [MLk,CLk,MLkadd,CLkadd] = Supervise(slk,clusters_gt,MLk,CLk);
        [MLk,CLk,qlid] = Propagation(MLk,CLk,MLkadd,CLkadd,K==2,false);
        if (idk==1)
            qlk(idk,:) = [qlid,sum(qlid)];
        else
            qlk(idk,:) = qlk(idk-1,:) + [qlid,sum(qlid)];
        end
    end
    qk = ((loop-1)*qk+qlk)/loop;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Affichage
    %x = NB_CONSTRAINT*linspace(0,MAX_LOOP-1,MAX_LOOP)';
    figure(1);
    plot(q(:,4),tps(:,2),'-k',q(:,4),tps(:,1),'--k','LineWidth',2);
    xlabel('# supervised pairs');
    ylabel('Propagation time (ms)');
    legend('Original matrix implementation','Optimized matrix implementation','Location','northwest');
    figure(2);
    plot(q(:,4),q(:,1),':k',q(:,4),q(:,2),'-.k',q(:,4),q(:,3),'--k',q(:,4),q(:,5),'-k','LineWidth',2);
    xlabel('# supervised pairs');
    ylabel('# propagated pairs');
    legend('1ère règle','2ème règle','3ème règle','Somme des propagations','Location','northwest');
    figure(3);
    clf;
    plot(q(:,4),q(:,1),':k',q(:,4),q(:,2),'-.k',q(:,4),q(:,3),'--k',q(:,4),q(:,5),'-k','LineWidth',2);
    hold on;
    plot(qs3(:,4),qs3(:,1),':r',qs3(:,4),qs3(:,2),'-.r',qs3(:,4),qs3(:,5),'-r');
    xlabel('# supervised pairs');
    ylabel('# propagated pairs');
    legend('1ère règle avec propagations 1, 2 et 3','2ème règle avec propagations 1, 2 et 3','3ème règle avec propagations 1, 2 et 3','Somme des propagations 1, 2 et 3','1ère règle avec propagations 1 et 2','2ème règle avec propagations 1 et 2','Somme des propagations 1 et 2','Location','northwest');
    figure(4);
    clf;
    plot(q(:,4),q(:,1),':k',q(:,4),q(:,2),'-.k',q(:,4),q(:,3),'--k',q(:,4),q(:,5),'-k','LineWidth',2);
    hold on;
    plot(qk(:,4),qk(:,1),':r',qk(:,4),qk(:,2),'-.r',qk(:,4),qk(:,3),'--r',qk(:,4),qk(:,5),'-r');
    xlabel('# supervised pairs');
    ylabel('# propagated pairs');
    legend('1ère règle','2ème règle','3ème règle','Somme des propagations','1ère règle avec contraintes liées','2ème règle avec contraintes liées','3ème règle avec contraintes liées','Somme avec contraintes liées','Location','northwest');
    drawnow;
end


